@@ -32,12 +32,12 @@ def_selector_args_handler(vector strvec)
 }
 
 static void
-def_iopolicy_handler(vector strvec)
+def_pgpolicy_handler(vector strvec)
 {
 	char * buff;
 
 	buff = set_value(strvec);
-	conf->default_iopolicy = get_pgpolicy_id(buff);
+	conf->default_pgpolicy = get_pgpolicy_id(buff);
 	free(buff);
 }
 
@@ -53,6 +53,12 @@ def_prio_callout_handler(vector strvec)
 	conf->default_getprio = set_value(strvec);
 }
 
+static void
+def_features_handler(vector strvec)
+{
+	conf->default_features = set_value(strvec);
+}
+
 /*
  * blacklist block handlers
  */
@@ -94,36 +100,34 @@ device_handler(vector strvec)
 static void
 vendor_handler(vector strvec)
 {
-	struct hwentry * hwe =
-		VECTOR_SLOT(conf->hwtable, VECTOR_SIZE(conf->hwtable) - 1);
+	struct hwentry * hwe = VECTOR_LAST_SLOT(conf->hwtable);
+
 	hwe->vendor = set_value(strvec);
 }
 
 static void
 product_handler(vector strvec)
 {
-	struct hwentry * hwe =
-		VECTOR_SLOT(conf->hwtable, VECTOR_SIZE(conf->hwtable) - 1);
+	struct hwentry * hwe = VECTOR_LAST_SLOT(conf->hwtable);
+
 	hwe->product = set_value(strvec);
 }
 
 static void
-hw_iopolicy_handler(vector strvec)
+hw_pgpolicy_handler(vector strvec)
 {
 	char * buff;
-	struct hwentry * hwe =
-		VECTOR_SLOT(conf->hwtable, VECTOR_SIZE(conf->hwtable) - 1);
+	struct hwentry * hwe = VECTOR_LAST_SLOT(conf->hwtable);
 
 	buff = set_value(strvec);
-	hwe->iopolicy = get_pgpolicy_id(buff);
+	hwe->pgpolicy = get_pgpolicy_id(buff);
 	free(buff);
 }
 
 static void
 hw_getuid_callout_handler(vector strvec)
 {
-	struct hwentry * hwe =
-		VECTOR_SLOT(conf->hwtable, VECTOR_SIZE(conf->hwtable) - 1);
+	struct hwentry * hwe = VECTOR_LAST_SLOT(conf->hwtable);
 
 	hwe->getuid = set_value(strvec);
 }
@@ -131,7 +135,7 @@ hw_getuid_callout_handler(vector strvec)
 static void
 hw_selector_handler(vector strvec)
 {
-	struct hwentry * hwe = VECTOR_SLOT(conf->hwtable, VECTOR_SIZE(conf->hwtable) - 1);
+	struct hwentry * hwe = VECTOR_LAST_SLOT(conf->hwtable);
 	
 	hwe->selector = set_value(strvec);
 }
@@ -140,7 +144,7 @@ static void
 hw_selector_args_handler(vector strvec)
 {
 	char * buff;
-	struct hwentry * hwe = VECTOR_SLOT(conf->hwtable, VECTOR_SIZE(conf->hwtable) - 1);
+	struct hwentry * hwe = VECTOR_LAST_SLOT(conf->hwtable);
 
 	buff = set_value(strvec);
 	hwe->selector_args = atoi(buff);
@@ -152,7 +156,7 @@ static void
 hw_path_checker_handler(vector strvec)
 {
 	char * buff;
-	struct hwentry * hwe = VECTOR_SLOT(conf->hwtable, VECTOR_SIZE(conf->hwtable) - 1);
+	struct hwentry * hwe = VECTOR_LAST_SLOT(conf->hwtable);
 
 	buff = set_value(strvec);
 	hwe->checker_index = get_checker_id(buff);
@@ -160,6 +164,30 @@ hw_path_checker_handler(vector strvec)
 	free(buff);
 }
 
+static void
+hw_features_handler(vector strvec)
+{
+	struct hwentry * hwe = VECTOR_LAST_SLOT(conf->hwtable);
+	
+	hwe->features = set_value(strvec);
+}
+
+static void
+hw_handler_handler(vector strvec)
+{
+	struct hwentry * hwe = VECTOR_LAST_SLOT(conf->hwtable);
+	
+	hwe->hwhandler = set_value(strvec);
+}
+
+static void
+prio_callout_handler(vector strvec)
+{
+	struct hwentry * hwe = VECTOR_LAST_SLOT(conf->hwtable);
+	
+	hwe->getprio = set_value(strvec);
+}
+
 /*
  * multipaths block handlers
  */
@@ -196,13 +224,13 @@ alias_handler(vector strvec)
 }
 
 static void
-mp_iopolicy_handler(vector strvec)
+mp_pgpolicy_handler(vector strvec)
 {
 	char * buff;
 	struct mpentry * mpe = VECTOR_SLOT(conf->mptable, VECTOR_SIZE(conf->mptable) - 1);
 
 	buff = set_value(strvec);
-	mpe->iopolicy = get_pgpolicy_id(buff);
+	mpe->pgpolicy = get_pgpolicy_id(buff);
 	free(buff);
 }
 
@@ -235,9 +263,10 @@ init_keywords(void)
 	install_keyword("udev_dir", &udev_dir_handler);
 	install_keyword("default_selector", &def_selector_handler);
 	install_keyword("default_selector_args", &def_selector_args_handler);
-	install_keyword("default_path_grouping_policy", &def_iopolicy_handler);
+	install_keyword("default_path_grouping_policy", &def_pgpolicy_handler);
 	install_keyword("default_getuid_callout", &def_getuid_callout_handler);
 	install_keyword("default_prio_callout", &def_prio_callout_handler);
+	install_keyword("default_features", &def_features_handler);
 	
 	install_keyword_root("devnode_blacklist", &blacklist_handler);
 	install_keyword("devnode", &devnode_handler);
@@ -247,11 +276,14 @@ init_keywords(void)
 	install_sublevel();
 	install_keyword("vendor", &vendor_handler);
 	install_keyword("product", &product_handler);
-	install_keyword("path_grouping_policy", &hw_iopolicy_handler);
+	install_keyword("path_grouping_policy", &hw_pgpolicy_handler);
 	install_keyword("getuid_callout", &hw_getuid_callout_handler);
 	install_keyword("path_selector", &hw_selector_handler);
 	install_keyword("path_selector_args", &hw_selector_args_handler);
 	install_keyword("path_checker", &hw_path_checker_handler);
+	install_keyword("features", &hw_features_handler);
+	install_keyword("hardware_handler", &hw_handler_handler);
+	install_keyword("prio_callout", &prio_callout_handler);
 	install_sublevel_end();
 
 	install_keyword_root("multipaths", &multipaths_handler);
@@ -259,7 +291,7 @@ init_keywords(void)
 	install_sublevel();
 	install_keyword("wwid", &wwid_handler);
 	install_keyword("alias", &alias_handler);
-	install_keyword("path_grouping_policy", &mp_iopolicy_handler);
+	install_keyword("path_grouping_policy", &mp_pgpolicy_handler);
 	install_keyword("path_selector", &mp_selector_handler);
 	install_keyword("path_selector_args", &mp_selector_args_handler);
 	install_sublevel_end();
