@@ -49,6 +49,219 @@
 #define argis(x) if (0 == strcmp (x, argv[i]))
 #define MATCH(x,y) 0 == strncmp (x, y, strlen(y))
 
+static struct hwentry *
+find_hw (char * vendor, char * product)
+{
+	int i;
+	struct hwentry * hwe;
+
+	vector_foreach_slot (conf->hwtable, hwe, i)
+		if (hwe->vendor && hwe->product &&
+		    strcmp(hwe->vendor, vendor) == 0 &&
+		    strcmp(hwe->product, product) == 0)
+			return hwe;
+	return NULL;
+}
+
+static struct mpentry *
+find_mp (char * wwid)
+{
+	int i;
+	struct mpentry * mpe;
+
+	vector_foreach_slot (conf->mptable, mpe, i)
+                if (mpe->wwid && strcmp(mpe->wwid, wwid) == 0)
+			return mpe;
+
+	return NULL;
+}
+
+/*
+ * selectors :
+ * traverse the configuration layers from most specific to most generic
+ * stop at first explicit setting found
+ */
+static int
+select_pgpolicy (struct multipath * mp)
+{
+	int i;
+	struct mpentry * mpe = NULL;
+	struct hwentry * hwe = NULL;
+	struct path * pp;
+	char pgpolicy_name[POLICY_NAME_SIZE];
+
+	pp = VECTOR_SLOT(mp->paths, 0);
+
+	if (conf->pgpolicy_flag > 0) {
+		mp->pgpolicy = conf->pgpolicy_flag;
+		get_pgpolicy_name(pgpolicy_name, mp->pgpolicy);
+		dbg("pgpolicy = %s (cmd line flag)", pgpolicy_name);
+		return 0;
+	}
+	vector_foreach_slot (conf->mptable, mpe, i) {
+		if (mpe->wwid && strcmp(mpe->wwid, mp->wwid) == 0 &&
+		    mpe->pgpolicy > 0) {
+			mp->pgpolicy = mpe->pgpolicy;
+			get_pgpolicy_name(pgpolicy_name, mp->pgpolicy);
+			dbg("pgpolicy = %s (LUN setting)", pgpolicy_name);
+			return 0;
+		}
+	}
+	vector_foreach_slot (conf->hwtable, hwe, i) {
+		if (hwe->vendor && hwe->product &&
+		    MATCH (pp->vendor_id, hwe->vendor) &&
+		    MATCH (pp->product_id, hwe->product) &&
+		    hwe->pgpolicy > 0) {
+			mp->pgpolicy = hwe->pgpolicy;
+			get_pgpolicy_name(pgpolicy_name, mp->pgpolicy);
+			dbg("pgpolicy = %s (controler setting)", pgpolicy_name);
+			return 0;
+		}
+	}
+	if (conf->default_pgpolicy > 0) {
+		mp->pgpolicy = conf->default_pgpolicy;
+		get_pgpolicy_name(pgpolicy_name, mp->pgpolicy);
+		dbg("pgpolicy = %s (config file default)", pgpolicy_name);
+		return 0;
+	}
+	mp->pgpolicy = FAILOVER;
+	get_pgpolicy_name(pgpolicy_name, FAILOVER);
+	dbg("pgpolicy = %s (internal default)", pgpolicy_name);
+	return 0;
+}
+
+static int
+select_selector (struct multipath * mp)
+{
+	int i;
+	struct mpentry * mpe = NULL;
+	struct hwentry * hwe = NULL;
+	struct path * pp;
+
+	pp = VECTOR_SLOT(mp->paths, 0);
+
+	mpe = find_mp(mp->wwid);
+
+	if (mpe && mpe->selector) {
+		mp->selector = mpe->selector;
+		mp->selector_args = mpe->selector_args;
+		dbg("selector = %s (LUN setting)", mp->selector);
+		dbg("selector_args = %s (LUN setting)", mp->selector_args);
+		return 0;
+	}
+	hwe = find_hw(pp->vendor_id, pp->product_id);
+
+	if (hwe && hwe->selector) {
+		mp->selector = hwe->selector;
+		mp->selector_args = hwe->selector_args;
+		dbg("selector = %s (controler setting)", mp->selector);
+		dbg("selector_args = %s (controler setting)", mp->selector_args);
+		return 0;
+	}
+	mp->selector = conf->default_selector;
+	mp->selector_args = conf->default_selector_args;
+	dbg("selector = %s (internal default)", mp->selector);
+	dbg("selector_args = %s (internal default)", mp->selector_args);
+	return 0;
+}
+
+static int
+select_features (struct multipath * mp)
+{
+	int i;
+	struct hwentry * hwe = NULL;
+	struct path * pp;
+
+	pp = VECTOR_SLOT(mp->paths, 0);
+	hwe = find_hw(pp->vendor_id, pp->product_id);
+
+	if (hwe && hwe->features) {
+		mp->features = hwe->features;
+		dbg("features = %s (controler setting)", mp->features);
+		return 0;
+	}
+	mp->features = conf->default_features;
+	dbg("features = %s (internal default)", mp->features);
+	return 0;
+}
+
+static int
+select_hwhandler (struct multipath * mp)
+{
+	int i;
+	struct hwentry * hwe = NULL;
+	struct path * pp;
+
+        pp = VECTOR_SLOT(mp->paths, 0);
+	hwe = find_hw(pp->vendor_id, pp->product_id);
+
+	if (hwe && hwe->hwhandler) {
+		mp->hwhandler = hwe->hwhandler;
+		dbg("hwhandler = %s (controler setting)", mp->hwhandler);
+		return 0;
+	}
+	mp->hwhandler = conf->default_hwhandler;
+	dbg("hwhandler = %s (internal default)", mp->hwhandler);
+	return 0;
+}
+
+static int
+select_checkfn(struct path *pp)
+{
+	char checker_name[CHECKER_NAME_SIZE];
+	int i;
+	struct hwentry * hwe = NULL;
+
+	hwe = find_hw(pp->vendor_id, pp->product_id);
+
+	if (hwe && hwe->checker_index > 0) {
+		get_checker_name(checker_name, hwe->checker_index);
+		dbg("path checker = %s (controler setting)", checker_name);
+		pp->checkfn = get_checker_addr(hwe->checker_index);
+		return 0;
+	}
+	pp->checkfn = &readsector0;
+	get_checker_name(checker_name, READSECTOR0);
+	dbg("path checker = %s (internal default)", checker_name);
+	return 0;
+}
+
+static int
+select_getuid (struct path * pp)
+{
+	int i;
+	struct hwentry * hwe = NULL;
+
+	hwe = find_hw(pp->vendor_id, pp->product_id);
+
+	if (hwe && hwe->getuid) {
+		pp->getuid = hwe->getuid;
+		dbg("getuid = %s (controler setting)", pp->getuid);
+		return 0;
+	}
+	pp->getuid = conf->default_getuid;
+	dbg("getuid = %s (internal default)", pp->getuid);
+	return 0;
+}
+
+static int
+select_getprio (struct path * pp)
+{
+	int i;
+	struct hwentry * hwe = NULL;
+
+	hwe = find_hw(pp->vendor_id, pp->product_id);
+
+	if (hwe && hwe->getprio) {
+		pp->getprio = hwe->getprio;
+		dbg("getprio = %s (controler setting)", pp->getprio);
+		return 0;
+	}
+	pp->getprio = conf->default_getprio;
+	dbg("getprio = %s (internal default)", pp->getprio);
+	return 0;
+}
+
 static int
 devt2devname (char *devname, char *devt)
 {
@@ -116,146 +329,150 @@ blacklist (char * dev) {
 	return 0;
 }
 
-static int
-select_checkfn(struct path *pp)
+static char *
+apply_format (char * string, int maxsize, struct path * pp)
 {
-	char checker_name[CHECKER_NAME_SIZE];
-	int i;
-	struct hwentry * hwe;
+	char * pos;
+	char * dst;
+	char * p;
+	char c;
+	int len;
+	int free;
 
-	/*
-	 * default checkfn
-	 */
-	pp->checkfn = &readsector0;
+	dst = zalloc(maxsize);
 
-	vector_foreach_slot (conf->hwtable, hwe, i) {
-		if (MATCH(hwe->vendor, pp->vendor_id) &&
-		    MATCH(hwe->product, pp->product_id) &&
-		    hwe->checker_index > 0) {
-			get_checker_name(checker_name, hwe->checker_index);
-			dbg("set %s path checker for %s",
-				checker_name, pp->dev);
-			pp->checkfn = get_checker_addr(hwe->checker_index);
-			return 0;
-		}
+	if (!dst)
+		return NULL;
+
+	p = dst;
+	pos = strchr(string, '%');
+	free = maxsize;
+
+	if (!pos)
+		return string;
+
+	len = (int) (pos - string) + 1;
+	free -= len;
+
+	if (free < 2)
+		return NULL;
+
+	snprintf(p, len, "%s", string);
+	p += len - 1;
+	pos++;
+
+	switch (*pos) {
+	case 'n':
+		len = strlen(pp->dev) + 1;
+		free -= len;
+
+		if (free < 2)
+			return NULL;
+
+		snprintf(p, len, "%s", pp->dev);
+		p += len - 1;
+		break;
+	case 'd':
+		len = strlen(pp->dev_t) + 1;
+		free -= len;
+
+		if (free < 2)
+			return NULL;
+
+		snprintf(p, len, "%s", pp->dev_t);
+		p += len - 1;
+		break;
+	default:
+		break;
 	}
-	get_checker_name(checker_name, READSECTOR0);
-	dbg("set %s path checker for %s", checker_name, pp->dev);
-	return 0;
+	pos++;
+
+	if (!*pos)
+		return dst;
+
+	len = strlen(pos) + 1;
+	free -= len;
+
+	if (free < 2)
+		return NULL;
+
+	snprintf(p, len, "%s", pos);
+	dbg("reformated callout = %s", dst);
+	return dst;
 }
 
 static int
-devinfo (struct path *curpath)
+devinfo (struct path *pp)
 {
 	int i;
 	struct hwentry * hwe;
-	char buff[100];
+	char * buff;
 	char prio[16];
 
+	dbg("===== path %s =====", pp->dev);
+
 	/*
 	 * fetch info available in sysfs
 	 */
-	if (sysfs_devinfo(curpath))
+	if (sysfs_devinfo(pp))
 		return 1;
 
 	/*
 	 * then those not available through sysfs
 	 */
-	get_serial(curpath->serial, curpath->dev_t);
-	curpath->claimed = get_claimed(curpath->dev_t);
+	get_serial(pp->serial, pp->dev_t);
+	dbg("serial = %s", pp->serial);
+	pp->claimed = get_claimed(pp->dev_t);
+	dbg("claimed = %i", pp->claimed);
 
 	/*
 	 * get path state, no message collection, no context
 	 */
-	select_checkfn(curpath);
-	curpath->state = checkpath(curpath->dev_t, curpath->checkfn,
-				   NULL, NULL);
-	dbg("path %s state : %i", curpath->dev, curpath->state);
+	select_checkfn(pp);
+	pp->state = pp->checkfn(pp->dev_t, NULL, NULL);
+	dbg("state = %i", pp->state);
 	
 	/*
 	 * get path prio
 	 */
-	if(safe_sprintf(buff, "%s /block/%s",
-			conf->default_getprio, curpath->dev)) {
-		fprintf(stderr, "buff too small\n");
-		exit(1);
-	}
+	select_getprio(pp);
+	buff = apply_format(pp->getprio, CALLOUT_MAX_SIZE, pp);
 
-	dbg("get prio callout :");
-	dbg("==================");
-
-	if (execute_program(buff, prio, 16) == 0)
-		curpath->priority = atoi(prio);
-	else {
+	if (!buff)
+		pp->priority = 1;
+	else if (execute_program(buff, prio, 16)) {
 		dbg("error calling out %s", buff);
-		curpath->priority = 1;
-	}
-	dbg("devinfo found prio : %u", curpath->priority);
+		pp->priority = 1;
+	} else
+		pp->priority = atoi(prio);
+
+	dbg("prio = %u", pp->priority);
 
 	/*
 	 * get path uid
 	 */
-	vector_foreach_slot (conf->hwtable, hwe, i) {
-		if (MATCH (curpath->vendor_id, hwe->vendor) &&
-		    MATCH (curpath->product_id, hwe->product)) {
-			/*
-			 * callout program
-			 */
-			dbg("get uid callout :");
-			dbg("=================");
-			if(safe_sprintf(buff, "%s /block/%s",
-			   hwe->getuid ? hwe->getuid : conf->default_getuid,
-			   curpath->dev)) {
-				fprintf(stderr, "buff too small\n");
-				exit(1);
-			}
-			if (execute_program(buff, curpath->wwid,
-			    WWID_SIZE) == 0) {
-				dbg("devinfo found uid : %s", curpath->wwid);
-				return 0;
-			}
-			dbg("error calling out %s", buff);
-			dbg("falling back to internal getuid function");
+	select_getuid(pp);
 
-			/*
-			 * fallback
-			 */
-			if (!get_evpd_wwid(curpath->dev_t, curpath->wwid)) {
-				dbg("devinfo found uid : %s", curpath->wwid);
-				return 0;
-			}
-			/*
-			 * no wwid : blank for safety
-			 */
-			dbg("unable to fetch a wwid : set to 0x0");
-			memset(curpath->wwid, 0, WWID_SIZE);
-			return 1;
-		}
-	}
-	dbg("devinfo out : no match ... apply defaults");
+	buff = apply_format(pp->getuid, CALLOUT_MAX_SIZE, pp);
 
-	/*
-	 * chances are we deal directly with disks here (no FC ctlr)
-	 * we need scsi_id for this fallback to work
-	 *
-	 * incidentaly, dealing with this case will make parallel SCSI
-	 * disks treated as 1-path multipaths, which is good : wider audience !
-	 */
-	if(safe_sprintf(buff, "%s /block/%s",
-			conf->default_getuid, curpath->dev)) {
-		fprintf(stderr, "buff too small\n");
-		exit(1);
-	}
+	if (!buff)
+		goto fallback;
 
-	dbg("default get uid callout :");
-	dbg("=========================");
-
-	if (execute_program(buff, curpath->wwid, WWID_SIZE) == 0) {
-		dbg("devinfo found uid : %s", curpath->wwid);
+	if (execute_program(buff, pp->wwid, WWID_SIZE) == 0) {
+		dbg("uid = %s (callout)", pp->wwid);
 		return 0;
 	}
-	dbg("error calling out %s", buff);
-	memset(curpath->wwid, 0, WWID_SIZE);
+
+	fallback:
+	if (!get_evpd_wwid(pp->dev_t, pp->wwid)) {
+		dbg("uid = %s (internal getuid)", pp->wwid);
+		return 0;
+	}
+	/*
+	 * no wwid : blank for safety
+	 */
+	dbg("uid = 0x0 (unable to fetch)");
+	memset(pp->wwid, 0, WWID_SIZE);
 	return 1;
 }
 
@@ -380,42 +597,44 @@ get_pathvec_sysfs (vector pathvec)
 static void
 print_path (struct path * pp, int style)
 {
-	if (style != SHORT)
+	if (style != SHORT && pp->wwid)
 		printf ("%s ", pp->wwid);
 	else
 		printf (" \\_");
 
-	printf ("(%i %i %i %i) ",
+	printf("(%i %i %i %i) ",
 	       pp->sg_id.host_no,
 	       pp->sg_id.channel,
 	       pp->sg_id.scsi_id,
 	       pp->sg_id.lun);
 
-	printf ("%s ", pp->dev);
+	if (pp->dev)
+		printf("%s ", pp->dev);
 
 	switch (pp->state) {
 	case PATH_UP:
-		printf ("[ready ] ");
+		printf("[ready ]");
 		break;
 	case PATH_DOWN:
-		printf ("[faulty] ");
+		printf("[faulty]");
 		break;
 	case PATH_SHAKY:
-		printf ("[shaky] ");
+		printf("[shaky ]");
 		break;
 	default:
-		dbg("undefined path state");
+		printf("[undef ]");
 		break;
 	}
-	printf ("(%s) ", pp->dev_t);
+	if (pp->dev_t)
+		printf("[%s]", pp->dev_t);
 
 	if (pp->claimed)
-		printf ("[claimed] ");
+		printf("[claimed]");
 
-	if (style != SHORT)
-		printf ("[%.16s]", pp->product_id);
+	if (style != SHORT && pp->product_id)
+		printf("[%.16s]", pp->product_id);
 
-	fprintf (stdout, "\n");
+	fprintf(stdout, "\n");
 }
 
 static void
@@ -446,35 +665,28 @@ print_all_mp (vector mp)
 	struct multipath * mpp;
 	struct path * pp = NULL;
 
-	fprintf (stdout, "#\n# all multipaths :\n#\n");
+	fprintf(stdout, "#\n# all multipaths :\n#\n");
 
 	vector_foreach_slot (mp, mpp, k) {
 		if (mpp->alias)
-			printf ("%s (%s)", mpp->alias, mpp->wwid);
+			printf("%s (%s)", mpp->alias, mpp->wwid);
 		else
-			printf ("%s", mpp->wwid);
+			printf("%s", mpp->wwid);
 
+		if (!mpp->paths) {
+			printf("\n");
+			continue;
+		}
 		pp = VECTOR_SLOT(mpp->paths, 0);
-		printf (" [%.16s]\n", pp->product_id);
+
+		if (pp->product_id)
+			printf(" [%.16s]\n", pp->product_id);
 
 		vector_foreach_slot (mpp->paths, pp, i)
-			print_path (pp, SHORT);
+			print_path(pp, SHORT);
 	}
 }
 
-static struct mpentry *
-find_mp (char * wwid)
-{
-	int i;
-	struct mpentry * mpe;
-
-	vector_foreach_slot (conf->mptable, mpe, i)
-                if (mpe->wwid && strcmp(mpe->wwid, wwid) == 0)
-			return mpe;
-
-	return NULL;
-}
-
 static void
 coalesce_paths (vector mp, vector pathvec)
 {
@@ -539,42 +751,6 @@ coalesce_paths (vector mp, vector pathvec)
 	}
 }
 
-static int
-dm_get_map(char * name, char * outparams)
-{
-	int r = 0;
-	struct dm_task *dmt;
-	void *next = NULL;
-	uint64_t start, length;
-	char *target_type = NULL;
-	char *params;
-	int cmd;
-
-	cmd = DM_DEVICE_TABLE;
-
-	if (!(dmt = dm_task_create(cmd)))
-		return 0;
-
-	if (!dm_task_set_name(dmt, name))
-		goto out;
-
-	if (!dm_task_run(dmt))
-		goto out;
-
-	/* Fetch 1st target */
-	next = dm_get_next_target(dmt, next, &start, &length,
-				  &target_type, &params);
-
-	if (safe_snprintf(outparams, PARAMS_SIZE, params))
-		goto out;
-
-	r = 1;
-
-	out:
-	dm_task_destroy(dmt);
-	return r;
-}
-
 static int
 dm_reinstate(char * mapname, char * path)
 {
@@ -613,50 +789,6 @@ dm_reinstate(char * mapname, char * path)
         return r;
 }
 
-static int
-select_selector (struct multipath * mp)
-{
-	int i;
-	struct mpentry * mpe;
-	struct hwentry * hwe;
-	struct path * pp;
-
-	pp = VECTOR_SLOT(mp->paths, 0);
-
-	/*
-	 * select the right path selector :
-	 * 1) set internal default
-	 * 2) override by controler wide settings
-	 * 3) override by LUN specific settings
-	 */
-
-	/* 1) set internal default */
-	mp->selector = conf->default_selector;
-	mp->selector_args = conf->default_selector_args;
-
-	/* 2) override by controler wide settings */
-	vector_foreach_slot (conf->hwtable, hwe, i) {
-		if (strcmp(hwe->vendor, pp->vendor_id) == 0 &&
-		    strcmp(hwe->product, pp->product_id) == 0) {
-			mp->selector = (hwe->selector) ?
-				   hwe->selector : conf->default_selector;
-			mp->selector_args = (hwe->selector_args) ?
-				   hwe->selector_args : conf->default_selector_args;
-		}
-	}
-
-	/* 3) override by LUN specific settings */
-	vector_foreach_slot (conf->mptable, mpe, i) {
-		if (mpe->wwid && strcmp(mpe->wwid, mp->wwid) == 0) {
-			mp->selector = (mpe->selector) ?
-				   mpe->selector : conf->default_selector;
-			mp->selector_args = (mpe->selector_args) ?
-				   mpe->selector_args : conf->default_selector_args;
-		}
-	}
-	return 0;
-}
-
 /*
  * Transforms the path group vector into a proper device map string
  */
@@ -672,7 +804,9 @@ assemble_map (struct multipath * mp)
 	p = mp->params;
 	freechar = sizeof(mp->params);
 	
-	shift = snprintf(p, freechar, "%i", VECTOR_SIZE(mp->pg));
+	shift = snprintf(p, freechar, "%s %s %i 1",
+			 mp->features, mp->hwhandler,
+			 VECTOR_SIZE(mp->pg));
 
 	if (shift >= freechar) {
 		fprintf(stderr, "mp->params too small\n");
@@ -713,20 +847,17 @@ static int
 setup_map (vector pathvec, struct multipath * mpp)
 {
 	struct path * pp;
-	int i, iopolicy;
-	char iopolicy_name[POLICY_NAME_SIZE];
-	int fd;
-	struct hwentry * hwe = NULL;
-	struct mpentry * mpe;
-	char * mapname;
-	char curparams[PARAMS_SIZE];
+	int i;
+	char * mapname = NULL;
+	char * curparams = NULL;
 	int op;
+	int r = 0;
 
 	/*
 	 * don't bother if devmap size is unknown
 	 */
 	if (mpp->size <= 0)
-		return 0;
+		return 1;
 
 	/*
 	 * don't bother if a constituant path is claimed
@@ -734,75 +865,17 @@ setup_map (vector pathvec, struct multipath * mpp)
 	 */
 	vector_foreach_slot (mpp->paths, pp, i)
 		if (pp->claimed)
-			return 0;
+			return 1;
 
 	pp = VECTOR_SLOT(mpp->paths, 0);
 
 	/*
-	 * iopolicy selection logic :
-	 *  1) set internal default
-	 *  2) override by config file default
-	 *  3) override by controler wide setting
-	 *  4) override by LUN specific setting
-	 *  5) cmd line flag has the last word
-	 */
-	dbg("iopolicy selector :");
-	dbg("===================");
-
-	/*
-	 * 1) set internal default
-	 */
-	iopolicy = FAILOVER;
-	get_pgpolicy_name(iopolicy_name, FAILOVER);
-	dbg("internal default)\tiopolicy = %s", iopolicy_name);
-
-	/*
-	 * 2) override by config file default
-	 */
-	if (conf->default_iopolicy > 0) {
-		iopolicy = conf->default_iopolicy;
-		get_pgpolicy_name(iopolicy_name, iopolicy);
-		dbg("config file default)\tiopolicy = %s", iopolicy_name);
-	}
-	
-	/*
-	 * 3) override by controler wide setting
-	 */
-	vector_foreach_slot (conf->hwtable, hwe, i) {
-		if (MATCH (pp->vendor_id, hwe->vendor) &&
-		    MATCH (pp->product_id, hwe->product) &&
-		    hwe->iopolicy > 0) {
-			iopolicy = hwe->iopolicy;
-			get_pgpolicy_name(iopolicy_name, iopolicy);
-			dbg("controler override)\tiopolicy = %s", iopolicy_name);
-		}
-	}
-
-	/*
-	 * 4) override by LUN specific setting
-	 */
-	vector_foreach_slot (conf->mptable, mpe, i) {
-		if (mpe->wwid && strcmp(mpe->wwid, mpp->wwid) == 0 &&
-		    mpe->iopolicy > 0) {
-			iopolicy = mpe->iopolicy;
-			get_pgpolicy_name(iopolicy_name, iopolicy);
-			dbg("lun override)\t\tiopolicy = %s", iopolicy_name);
-		}
-	}
-
-	/*
-	 * 5) cmd line flag has the last word
-	 */
-	if (conf->iopolicy_flag > 0) {
-		iopolicy = conf->iopolicy_flag;
-		get_pgpolicy_name(iopolicy_name, iopolicy);
-		dbg("cmd flag override)\tiopolicy = %s", iopolicy_name);
-	}
-
-	/*
-	 * select the appropriate path group selector and args
+	 * properties selectors
 	 */
+	select_pgpolicy(mpp);
 	select_selector(mpp);
+	select_features(mpp);
+	select_hwhandler(mpp);
 
 	/*
 	 * layered map computation :
@@ -815,26 +888,26 @@ setup_map (vector pathvec, struct multipath * mpp)
 	/*
 	 * 1) & 2)
 	 */
-	group_by_status(mpp, PATH_DOWN);
-	group_by_status(mpp, PATH_SHAKY);
+	//group_by_status(mpp, PATH_DOWN);
+	//group_by_status(mpp, PATH_SHAKY);
 
 	/*
 	 * 3) apply selected grouping policy
 	 */
-	if (iopolicy == MULTIBUS)
+	if (mpp->pgpolicy == MULTIBUS)
 		one_group (mpp);
 
-	if (iopolicy == FAILOVER)
+	if (mpp->pgpolicy == FAILOVER)
 		one_path_per_group (mpp);
 
-	if (iopolicy == GROUP_BY_SERIAL)
+	if (mpp->pgpolicy == GROUP_BY_SERIAL)
 		group_by_serial (mpp);
 
-	if (iopolicy == GROUP_BY_PRIO)
+	if (mpp->pgpolicy == GROUP_BY_PRIO)
 		group_by_prio (mpp);
 
 	if (mpp->pg == NULL) {
-		dbg("iopolicy failed to produce a ->pg vector");
+		dbg("pgpolicy failed to produce a ->pg vector");
 		return 1;
 	}
 
@@ -870,26 +943,26 @@ setup_map (vector pathvec, struct multipath * mpp)
 	 * last chance to quit before touching the devmaps
 	 */
 	if (conf->dry_run)
-		return 1;
+		return 0;
 
 	/*
 	 * device mapper creation or updating
 	 * here we know we'll have garbage on stderr from
 	 * libdevmapper. so shut it down temporarily.
 	 */
-	fd = dup(2);
-	close(2);
+	dm_log_init_verbose(0);
 
 	if (op == DM_DEVICE_RELOAD) {
 		if (conf->dev && filepresent(conf->dev) &&
-		    dm_get_map(mapname, curparams) &&
+		    dm_get_map(mapname, NULL, &curparams) &&
 		    0 == strncmp(mpp->params, curparams, strlen(mpp->params))) {
 	                pp = zalloc(sizeof(struct path));
 	                basename(conf->dev, pp->dev);
 
-	                if (devinfo(pp))
-	                        return 1;
-
+	                if (devinfo(pp)) {
+				r = 1;
+	                        goto out;
+			}
 			dm_reinstate(mapname, pp->dev_t);
 	                free(pp);
 
@@ -902,7 +975,7 @@ setup_map (vector pathvec, struct multipath * mpp)
 			goto out;
 	}
 
-	if (dm_addmap(op, mapname, DM_TARGET, mpp->params, mpp->size)) {
+	if (!dm_addmap(op, mapname, DM_TARGET, mpp->params, mpp->size)) {
 		dm_simplecmd(DM_DEVICE_REMOVE, mapname);
 		goto out;
 	}
@@ -911,12 +984,59 @@ setup_map (vector pathvec, struct multipath * mpp)
 		if (!dm_simplecmd(DM_DEVICE_RESUME, mapname))
 			goto out;
 
-	return 1;
+	out:
+	dm_log_init_verbose(1);
+	free(curparams);
+	return r;
+}
+
+static int
+dm_get_maps (vector mp, char * type)
+{
+	struct multipath * mpp;
+	int r = 0;
+	struct dm_task *dmt;
+	struct dm_names *names;
+	unsigned next = 0;
+	unsigned long length;
+	char *params;
+
+	if (!(dmt = dm_task_create (DM_DEVICE_LIST)))
+		return 0;
+
+	if (!dm_task_run (dmt))
+		goto out;
+
+	if (!(names = dm_task_get_names (dmt)))
+		goto out;
+
+	if (!names->dev)
+		goto out;
+
+	do {
+		if (dm_type(names->name, DM_TARGET)) {
+			dm_get_map(names->name, &length, &params);
+			mpp = zalloc(sizeof(struct multipath));
+
+			if (!mpp) {
+				r = 1;
+				goto out;
+			}
+			mpp->size = length;
+			strncat(mpp->wwid, names->name, WWID_SIZE);
+			strncat(mpp->params, params, PARAMS_SIZE);
+
+			vector_alloc_slot(mp);
+			vector_set_slot(mp, mpp);
+			mpp = NULL;
+		}
+                next = names->next;
+                names = (void *) names + next;
+	} while (next);
 
 	out:
-	dup2(fd, 2);
-	close(fd);
-	return 0;
+	dm_task_destroy (dmt);
+	return r;
 }
 
 static void
@@ -976,7 +1096,7 @@ setup_default_blist (vector blist)
 	snprintf (hwe->vendor, SCSI_VENDOR_SIZE, b); \
 	hwe->product = zalloc (SCSI_PRODUCT_SIZE * sizeof(char)); \
 	snprintf (hwe->product, SCSI_PRODUCT_SIZE, c); \
-	hwe->iopolicy = d; \
+	hwe->pgpolicy = d; \
 	hwe->getuid = e; \
 	vector_alloc_slot(a); \
 	vector_set_slot(a, hwe);
@@ -1027,7 +1147,7 @@ usage (char * progname)
 		"\t-D maj:min\tlimit scope to the device's multipath\n" \
 		"\t\t\t(major:minor device reference)\n"
 		"\t-S\t\tinhibit signal sending to multipathd\n"
-		"\n" \
+		"\t-F\t\tflush all multipath device maps\n" \
 		"\t-p policy\tforce all maps to specified policy :\n" \
 		"\t   failover\t\t1 path per priority group\n" \
 		"\t   multibus\t\tall paths in 1 priority group\n" \
@@ -1078,7 +1198,7 @@ int try_lock (char * file)
 int
 main (int argc, char *argv[])
 {
-	vector mp;
+	vector mp, curmp;
 	vector pathvec;
 	struct multipath * mpp;
 	int k;
@@ -1086,7 +1206,7 @@ main (int argc, char *argv[])
 	extern char *optarg;
 	extern int optind;
 
-	if (dm_prereq(DM_TARGET, 0, 0, 0)) {
+	if (dm_prereq(DM_TARGET, 1, 0, 3)) {
 		dbg("device mapper prerequisites not met");
 		exit(1);
 	}
@@ -1108,20 +1228,23 @@ main (int argc, char *argv[])
 	/*
 	 * internal defaults
 	 */
+	conf->list = 0;
 	conf->dry_run = 0;		/* 1 == Do not Create/Update devmaps */
-	conf->verbosity = 1;		/* 1 == Print mp names */
-	conf->iopolicy_flag = 0;	/* do not override defaults */
+	conf->verbosity = 1;
+	conf->pgpolicy_flag = 0;	/* do not override defaults */
 	conf->signal = 1;		/* 1 == Send a signal to multipathd */
 	conf->dev = NULL;
 	conf->devt = NULL;
 	conf->default_selector = NULL;
 	conf->default_selector_args = 0;
-	conf->default_iopolicy = 0;
+	conf->default_pgpolicy = 0;
 	conf->mptable = NULL;
 	conf->hwtable = NULL;
 	conf->blist = NULL;
+	conf->default_features = NULL;
+	conf->default_hwhandler = NULL;
 
-	while ((arg = getopt(argc, argv, ":qdSi:v:p:D:")) != EOF ) {
+	while ((arg = getopt(argc, argv, ":qdlFSi:v:p:D:")) != EOF ) {
 		switch(arg) {
 		case 1: printf("optarg : %s\n",optarg);
 			break;
@@ -1136,18 +1259,26 @@ main (int argc, char *argv[])
 			conf->dry_run = 1;
 			conf->signal = 0;
 			break;
+		case 'F':
+			dm_flush_maps(DM_TARGET);
+			goto out;
+			break;
+		case 'l':
+			conf->list = 1;
+			conf->signal = 0;
+			break;
 		case 'S':
 			conf->signal = 0;
 			break;
 		case 'p':
 			if (strcmp(optarg, "failover") == 0)
-				conf->iopolicy_flag = FAILOVER;
+				conf->pgpolicy_flag = FAILOVER;
 			else if (strcmp(optarg, "multibus") == 0)
-				conf->iopolicy_flag = MULTIBUS;
+				conf->pgpolicy_flag = MULTIBUS;
 			else if (strcmp(optarg, "group_by_serial") == 0)
-				conf->iopolicy_flag = GROUP_BY_SERIAL;
+				conf->pgpolicy_flag = GROUP_BY_SERIAL;
 			else if (strcmp(optarg, "group_by_prio") == 0)
-				conf->iopolicy_flag = GROUP_BY_PRIO;
+				conf->pgpolicy_flag = GROUP_BY_PRIO;
 			else
 			{
 				if (optarg)
@@ -1175,6 +1306,24 @@ main (int argc, char *argv[])
 		strncpy(conf->dev, argv[optind], FILE_NAME_SIZE);
 	}
 
+	/*
+	 * allocate the two core vectors to store paths and multipaths
+	 */
+	mp = vector_alloc();
+	curmp = vector_alloc();
+	pathvec = vector_alloc();
+
+	if (!mp || !curmp || !pathvec) {
+		fprintf(stderr, "can not allocate memory\n");
+		exit(1);
+	}
+
+	if (conf->list) {
+		dm_get_maps(curmp, DM_TARGET);
+		print_all_mp(curmp);
+		goto out;
+	}
+
 	/*
 	 * read the config file
 	 */
@@ -1204,17 +1353,15 @@ main (int argc, char *argv[])
 	if (conf->default_getuid == NULL)
 		conf->default_getuid = DEFAULT_GETUID;
 
+	if (conf->default_features == NULL)
+		conf->default_features = DEFAULT_FEATURES;
+
+	if (conf->default_hwhandler == NULL)
+		conf->default_hwhandler = DEFAULT_HWHANDLER;
+
 	/*
-	 * allocate the two core vectors to store paths and multipaths
+	 * get a path list and group them as multipaths
 	 */
-	mp = vector_alloc();
-	pathvec = vector_alloc();
-
-	if (mp == NULL || pathvec == NULL) {
-		fprintf(stderr, "can not allocate memory\n");
-		exit(1);
-	}
-
 	if (get_pathvec_sysfs(pathvec))
 		exit(1);
 
@@ -1226,7 +1373,7 @@ main (int argc, char *argv[])
 	coalesce_paths(mp, pathvec);
 
 	/*
-	 * may be conf->dev is a mapname
+	 * conf->dev can be a mapname
 	 * if so, only reconfigure this map
 	 */
 	vector_foreach_slot (mp, mpp, k) {
@@ -1258,9 +1405,10 @@ out:
 		signal_daemon();
 	
 	/*
-	 * free allocs
+	 * free allocs : FIXME, need a freemp() or nothing
 	 */
 	free(mp);
+	free(curmp);
 	free(pathvec);
 
 	exit(0);
