@@ -30,6 +30,7 @@
 #define PARAMS_SIZE	256
 #define FILE_NAME_SIZE	256
 #define DEV_T_SIZE	32
+#define CALLOUT_MAX_SIZE	128
 #define DEF_TIMEOUT	60000
 #define DM_TARGET	"multipath"
 #define PIDFILE		"/var/run/multipathd.pid"
@@ -38,6 +39,8 @@
 #define DEFAULT_GETUID	"/sbin/scsi_id -g -u -s"
 #define DEFAULT_UDEVDIR	"/udev"
 #define DEFAULT_SELECTOR	"round-robin"
+#define DEFAULT_FEATURES	"0"
+#define DEFAULT_HWHANDLER	"0"
 
 /*
  * SCSI strings sizes
@@ -84,26 +87,30 @@ struct path {
 	int state;
 	unsigned int priority;
 	int claimed;
-	int (*checkfn) (char *, char *, void *);
+	char * getuid;
+	char * getprio;
+	int (*checkfn) (char *, char *, void **);
 };
 
 struct multipath {
 	char wwid[WWID_SIZE];
 	char * alias;
-	int iopolicy;
+	int pgpolicy;
 	unsigned long size;
 	vector paths;
 	vector pg;
 	char params[PARAMS_SIZE];
 	char * selector;
 	int selector_args;
+	char * features;
+	char * hwhandler;
 };
 
 /* Build version */
 #define PROG    "multipath"
 
-#define VERSION_CODE 0x000307
-#define DATE_CODE    0x050b04
+#define VERSION_CODE 0x000308
+#define DATE_CODE    0x260b04
 
 #define MULTIPATH_VERSION(version)	\
 	(version >> 16) & 0xFF,		\
